module Main where

{-
Monoid:

class Monoid m where
  mempty  :: m
  mappend :: m -> m -> m 

Um valor do tipo `m` pode ser reescrito como uma função:

mempty :: () -> m 

e a assinatura m -> m -> m é equivalente a (m,m) -> m 

class Monoid m where
  mempty  :: () -> m
  mappend :: (m,m) -> m

monoid :: (() -> m, (m,m) -> m)

(m^1, m^(m*m)) = m * m^(m*m) = m ^ (m*m + 1) 

toda função:

monoid = m*m + 1 -> m

é candidata a ser um monoid!

m*m + 1 é um produto de m's somado com um valor unitário! (ex.: Maybe (m,m))
E sabemos que essas estruturas algébricas são Functores! Podemos generalizar o que fizemos acima com:

F m -> m

Com isso temos o conceito de Functor Algebra que consiste em:

- Um Functor F
- Um tipo a
- Uma função F a -> a

-}

type Algebra f a = f a -> a 

data MonF a = MEmpty | MAppend a a

evalM :: Algebra MonF Integer
evalM MEmpty = 0
evalM (MAppend x y) = x + y

{- 
Não podemos fazer algo como:

MAppend (MAppend 1 2) 3

m1 = MAppend 1 2 :: MonF Integer
m2 = MAppend m1 3 :: MonF Integer ou MonF (MonF Integer) ?

type MonF1 a = MonF (MonF a)
type MonF2 a = Monf (MonF1 a)
type MonF(N+1) a = MonF (MonFN a)

Se N tende a infinito, em certo ponto temos que:

-}

newtype Fix f = Fix (f (Fix f))

-- o valor de um Fix f é f (Fix f)!
-- ou seja, se aplicarmos o f em Fix f temos como resultado o próprio Fix f!!
-- isso é chamado de ponto fixo

unFix :: Fix f -> f (Fix f)
unFix (Fix x) = x

{- 
com isso podemos fazer:

evalFix :: Fix MonF -> Integer
evalFix (Fix MEmpty) = 0
evalFix (Fix (MAppend x y)) = evalFix x + evalFix y

só que isso nos traz outros problemas, além de complicar demais nosso código!

Podemos usar o fato de que `f` é naturalmente um functor para criar funções auxiliares!
Vamos fazer um exemplo com os números naturais:
-}

data NatF a = ZeroF | SuccF a

{-
o ponto fixo é:

data Nat = NatF (Fix NatF)
         = ZeroF | SuccF (Fix NatF)
         = ZeroF | SuccF (NatF (Fix NatF))
         = ZeroF | SuccF NatF
-}

data Nat = Zero | Succ Nat 

instance Functor NatF where
  fmap f ZeroF     = ZeroF
  fmap f (SuccF x) = SuccF (f x)

-- NatF Integer -> Integer 
alg :: Algebra NatF Integer
alg ZeroF     = 0
alg (SuccF x) = x+1

dois :: Fix NatF
dois = Fix (SuccF (Fix (SuccF (Fix ZeroF))))

{-
unFix dois = SuccF (Fix (SuccF (Fix ZeroF)))
 fmap unfix
           = SuccF (SuccF (Fix ZeroF))
 fmap (fmap unfix)
           = SuccF (SuccF Zerof)
 fmap (fmap alg)
           = SuccF (SuccF 0)
 fmap alg 
           = SuccF 1
 alg
           = 2

o que queremos fazer é:

g alg = alg $ fmap g $ unFix natF 
       = (alg . fmap (g alg) . unFix) natF
       
-}

cata :: Functor f => (f a -> a) -> Fix f -> a
cata alg = alg . fmap (cata alg) . unFix

fibAlg :: NatF (Int, Int) -> (Int, Int)
fibAlg ZeroF          = (0, 1)
fibAlg (SuccF (m, n)) = (n, m+n)

fib = cata fibAlg

n2nat :: Int -> Fix NatF
n2nat 0 = Fix ZeroF
n2nat n = Fix (SuccF (n2nat (n-1)))

-- >>> fib (n2nat 10)
-- (55,89)

type CoAlgebra f a = a -> f a

ana :: Functor f => (a -> f a) -> a -> Fix f
ana coalg = Fix . fmap (ana coalg) . coalg

n2nat' :: CoAlgebra NatF Int
n2nat' 0 = ZeroF
n2nat' n = SuccF (n-1)

-- >>> fib (ana n2nat' 10)
-- (55,89)

{-
A parte interessante disso é que podemos combinar facilmente o anamorfismo e catamorfismo para
criar um processo de criação de dados (cata) e consumo de dados (ana) que não armazena nenhuma estrutura intermediária. 
-}

hylo :: Functor f => Algebra f a -> CoAlgebra f b -> b -> a
hylo alg coalg = alg . fmap (hylo alg coalg) . coalg

data ListF e a = NilF | ConsF e a

instance Functor (ListF e) where
  fmap f NilF        = NilF
  fmap f (ConsF e a) = ConsF e (f a)

lenAlg :: Algebra (ListF e) Int
lenAlg NilF        = 0
lenAlg (ConsF e n) = n + 1

sumAlg :: Algebra (ListF Integer) Integer
sumAlg NilF        = 0
sumAlg (ConsF e n) = e + n

listFibCo :: CoAlgebra (ListF Integer) (Integer, Integer, Int)
listFibCo (m, n, 0) = NilF
listFibCo (m, n, c) = ConsF (m+n) (n, m+n, c-1)

-- >>> hylo sumAlg listFibCo (0, 1, 15)
-- 2582


main :: IO ()
main = putStrLn "hello world"
